// Copyright (c) Lawrence Livermore National Security, LLC and other Conduit
// Project developers. See top-level LICENSE AND COPYRIGHT files for dates and
// other details. No copyright assignment is required to contribute to Conduit.

//-----------------------------------------------------------------------------
///
/// file: conduit_fmt.hpp
///
//-----------------------------------------------------------------------------

#ifndef CONDUIT_FMT_HPP
#define CONDUIT_FMT_HPP

#define FMT_EXCEPTIONS 1
#define FMT_HEADER_ONLY 1

// Include all fmt header files
#include "format.h"
#include "ostream.h"
#include "printf.h"

#endif

