// Copyright (c) Lawrence Livermore National Security, LLC and other Conduit
// Project developers. See top-level LICENSE AND COPYRIGHT files for dates and
// other details. No copyright assignment is required to contribute to Conduit.

//-----------------------------------------------------------------------------
///
/// file: conduit.hpp
///
//-----------------------------------------------------------------------------

#ifndef CONDUIT_HPP
#define CONDUIT_HPP


//-----------------------------------------------------------------------------
// -- includes for the public conduit interface -- 
//-----------------------------------------------------------------------------

#include "conduit_core.hpp"
#include "conduit_error.hpp"
#include "conduit_endianness.hpp"
#include "conduit_data_type.hpp"
#include "conduit_data_array.hpp"
#include "conduit_schema.hpp"
#include "conduit_node.hpp"
#include "conduit_generator.hpp"
#include "conduit_utils.hpp"
#include "conduit_data_accessor.hpp"

#endif

