#  tests for conduit-0.8.0-py39h53d6734_0 (this is a generated file);
print('===== testing package: conduit-0.8.0-py39h53d6734_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import conduit
import conduit.blueprint
import conduit.relay
import conduit.relay.io

print(conduit.about())
print(conduit.blueprint.about())
print(conduit.relay.about())
print(conduit.relay.io.about())
#  --- run_test.py (end) ---

print('===== conduit-0.8.0-py39h53d6734_0 OK =====');
print("import: 'conduit'")
import conduit

print("import: 'conduit.relay'")
import conduit.relay

print("import: 'conduit.blueprint'")
import conduit.blueprint

