// Copyright (c) Lawrence Livermore National Security, LLC and other Conduit
// Project developers. See top-level LICENSE AND COPYRIGHT files for dates and
// other details. No copyright assignment is required to contribute to Conduit.

//-----------------------------------------------------------------------------
///
/// file: conduit_relay_hdf5.hpp
///
//-----------------------------------------------------------------------------

#ifndef CONDUIT_RELAY_HDF5_HPP
#define CONDUIT_RELAY_HDF5_HPP

// NOTE: This file is provided for backward compatibility.
#pragma message("The conduit_relay_hdf5.hpp header file is deprecated. Include conduit_relay_io_hdf5.hpp instead.")

#include "conduit_relay_io_hdf5.hpp"

#endif

